/*
 *
 * demos/misc/disbonjour.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/utils/AppUtils.H>
#include <nucleo/core/ReactiveEngine.H>
#include <nucleo/network/rdv/ServiceAnnouncer.H>
#include <nucleo/network/rdv/ServiceFinder.H>
#include <nucleo/image/source/ImageSource.H>
#include <nucleo/image/sink/nserverImageSink.H>

#include <iostream>
#include <stdexcept>
#include <sstream>

using namespace nucleo ;

class ImageViewer : public ReactiveObject {

protected:

  ImageSource *source ;
  Image image ;
  ImageSink *sink ;
  bool delete_source, delete_sink ;

  void react(Observable *obs) {
    if (source->getNextImage(&image))
	 sink->handle(&image) ;
  }

private:

  void init(void) {
    subscribeTo(sink) ;
    sink->start() ;
    subscribeTo(source) ;
    source->start() ;
  }

public:

  ImageViewer(std::string src, std::string dst) {
    sink = ImageSink::create(dst) ;
    delete_sink = true ;
    source = ImageSource::create(src) ;
    delete_source = true ;
    init() ;
  }

  ImageViewer(ImageSource *src, ImageSink *dst, bool dSrc=true, bool dDst=true) {
    source = src ;
    sink = dst ;
    delete_source = dSrc ;
    delete_sink = dDst ;
    init() ;
  }

  bool isActive(void) {
    return (source->isActive() && sink->isActive()) ;
  }

  ~ImageViewer(void) {
    unsubscribeFrom(source) ;
    if (delete_source) delete source ;
    unsubscribeFrom(sink) ;
    if (delete_sink) delete sink ;
  }

} ;


class NiceGuy : public ReactiveObject {

protected:

  std::string name ;
  ImageSource *source ;
  Image image ;
  nserverImageSink *server ;
  ServiceAnnouncer *announcer ;

  ServiceFinder::EventQueue queue ;
  ServiceFinder *finder ;
  std::map<std::string,ImageViewer*> friends ;

  void react(Observable *obs) {
    if (source->getNextImage(&image))
	 server->handle(&image) ;

    if (queue.size()) {
	 ServiceFinder::Event *event = queue.front() ;
	 queue.pop() ;

	 std::map<std::string,ImageViewer*>::iterator f = friends.find(event->service.name) ;
	 if (f!=friends.end()) {
	   ImageViewer *viewer = (*f).second ;
	   if (viewer) {
		delete viewer ;
		friends[event->service.name] = 0 ;
	   }
	   if (event->event==ServiceFinder::Event::FOUND) {
		std::stringstream sourceURI, sinkURI ;
		sourceURI << "http://" << event->service.ipaddr << ":" << event->service.port ;
		std::string uri = sourceURI.str() ;
		std::cerr << "found " << event->service.name << " (" << uri << ")" << std::endl ;
		sinkURI << "glwindow:?title=" << event->service.name << "->" << name ;
		friends[event->service.name] = new ImageViewer(uri, sinkURI.str()) ;
	   } else if (event->event==ServiceFinder::Event::LOST) {
		std::cerr << "lost " << event->service.name << std::endl ;
	   }
	 } else {
	   std::cerr << "I don't care about " << event->service.name << std::endl ;
	 }

	 delete event ;
    }
  }

public:

  NiceGuy(char *src, char *n, char *type, int port) {
    name = n ;

    server = new nserverImageSink(port) ;
    if (!server) throw std::runtime_error("Unable to create server") ;
    subscribeTo(server) ;
    server->start() ;

    source = ImageSource::create(src) ;
    if (!source) throw std::runtime_error("Unable to create image source") ;
    subscribeTo(source) ;
    source->start() ;

    finder = ServiceFinder::create(type, &queue) ;
    if (!finder) throw std::runtime_error("Unable to create ServiceFinder") ;
    subscribeTo(finder) ;

    announcer = ServiceAnnouncer::create(n, type, server->getPortNumber()) ;
    if (!announcer) throw std::runtime_error("Unable to create ServiceAnnouncer") ;
  }

  void lookFor(std::string name) {
    std::map<std::string,ImageViewer*>::iterator i = friends.find(name) ;
    if (i!=friends.end()) return ;

    friends[name] = 0 ;
  }

  ~NiceGuy(void) {
    unsubscribeFrom(source) ;
    delete source ;
    unsubscribeFrom(server) ;
    delete server ;
    unsubscribeFrom(finder) ;
    delete finder ;
    delete announcer ;

    // FIXME: we should also delete the image sinks associated to our
    // friends...
  }

} ;

// ---------------------------------------------------------------------------

int
main(int argc, char **argv) {

  try {

    char *SOURCE = 0 ;
    char *NAME = "disbonjour" ;
    char *TYPE = "_http._tcp." ;
    int PORT = 0 ;

    int firstArg = parseCommandLine(argc, argv, "i:n:t:p:", "sssi",
							 &SOURCE, &NAME, &TYPE, &PORT) ;

    if (firstArg<0) {
	 std::cerr << std::endl << argv[0] ;
	 std::cerr << " [-i source] [-n name] peer1 ... peern" << std::endl ;
	 std::cerr << std::endl << std::endl ;
	 exit(1) ;
    }

    if (!SOURCE) SOURCE = getenv("NUCLEO_SRC1") ;

    NiceGuy ng(SOURCE, NAME, TYPE, PORT) ;
    for (int i=firstArg; i<argc; ++i) ng.lookFor(argv[i]) ;

    ReactiveEngine::run() ;

    return 0 ;
  } catch (std::runtime_error e) {
    std::cerr << "Runtime error: " << e.what() << std::endl ;
  } catch (...) {
    std::cerr << "Unknown error" << std::endl ;
  }

  return -1 ;
}
