/*
 *
 * nucleo/image/source/qtImgImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/qtImgImageSource.H>
#include <nucleo/image/encoding/Conversion.H>

#include <QuickTime/ImageCompression.h>

#include <stdexcept>

namespace nucleo {

  // --------------------------------------------------------------------------

  qtImgImageSource::qtImgImageSource(Image::Encoding encoding, const URI &uri) {
    _filename = uri.isOpaque() ? uri.getOpaque() : uri.getPath() ;
    _encoding = encoding ;
    _active = false ;

    ptrDataRefRec.dataLength = getFileSize(_filename.c_str()) ;
    ptrDataRefRec.data = (void *)Image::AllocMem(ptrDataRefRec.dataLength) ;
    readFromFile(_filename.c_str(), (unsigned char *)ptrDataRefRec.data, ptrDataRefRec.dataLength) ;

    Handle dataRef ;
    OSErr error = PtrToHand(&ptrDataRefRec, &dataRef, sizeof(PointerDataRefRecord)) ;
    if (error!=noErr)
	 throw std::runtime_error("qtImgImageSource: PtrToHand failed") ;

    error = GetGraphicsImporterForDataRef(dataRef, PointerDataHandlerSubType, &gi) ; 
    if (error!=noErr) {
	 DisposeHandle(dataRef) ;
	 Image::FreeMem((unsigned char **)&(ptrDataRefRec.data)) ;
	 std::string msg = "qtImgImageSource: GetGraphicsImporterForDataRef failed for " + _filename ;
	 throw std::runtime_error(msg) ;
    }
  }

  qtImgImageSource::~qtImgImageSource(void) {
    CloseComponent(gi) ;
    Image::FreeMem((unsigned char **)&(ptrDataRefRec.data)) ;
  }

  bool qtImgImageSource::start(void) {
    if (_active) return false ;

    // -----------------------------------

    ImageDescriptionHandle imageDescH = NULL ;
    ComponentResult result = GraphicsImportGetImageDescription(gi, &imageDescH);
    if (result!=noErr || imageDescH==NULL) {
	 std::cerr << "qtImgImageSource: GraphicsImportGetImageDescription failed (" << result << ")" << std::endl ;
	 return false ;
    }

    _image.prepareFor((*imageDescH)->width, (*imageDescH)->height, Image::ARGB) ;

    GWorldPtr gWorld ;
    Rect boundsRect = {0, 0,
				   (*imageDescH)->height, (*imageDescH)->width} ;
    OSErr error = NewGWorldFromPtr(&gWorld, k32ARGBPixelFormat,
							&boundsRect, NULL, NULL, 0, 
							(char *)_image.getData(),
							(*imageDescH)->width*4);
    if (error!=noErr) {
	 std::cerr << "qtImgImageSource: NewGWorldFromPtr failed (" << error << ")" << std::endl ;
	 return false ;
    }

    result = GraphicsImportSetGWorld(gi, gWorld, NULL) ;
    if (result!=noErr) {
	 std::cerr << "qtImgImageSource: GraphicsImportSetGWorld failed (" << result << ")" << std::endl ;
	 return false ;
    }

    result = GraphicsImportDraw(gi) ;
    if (result!=noErr) {
	 std::cerr << "qtImgImageSource: GraphicsImportDraw failed (" << result << ")" << std::endl ;
	 return false ;
    }

    DisposeHandle((Handle)imageDescH) ;
    DisposeGWorld(gWorld) ;

    // -----------------------------------

    convertImage(&_image, _encoding) ;
    
    _active = true ;
    _chrono.start() ;
    _chrono.tick() ;
    notifyObservers() ;

    return true ;
  }

  bool qtImgImageSource::getNextImage(Image *img) {
    if (!_active) return false ;
    img->linkDataFrom(_image) ;
    if (_pendingNotifications<1) stop() ;
    return true ;
  }

  bool qtImgImageSource::stop(void) {
    if (!_active) return false ;
    _chrono.stop() ;
    _active = false ;
    return true ;
  }

  // --------------------------------------------------------------------------

}
